@@ -1077,7 +1077,7 @@ class ParametersLEDET:
         time_vector_params[0] = 0
         time_vector_params[1] = 0.010
         time_vector_params[-1] = t_PC_LUT[-1]
-        time_vector_params[-2] = 0.5
+        time_vector_params[-2] = timeStep
         time_vector_params[-3] = time_vector_params[-4]+timeStep
         self.Options.time_vector_params = time_vector_params
 
