@@ -3,12 +3,15 @@ import matplotlib.pyplot as plt
 import matplotlib.cm as cm
 from matplotlib import colors as mcolors
 import h5py
+import copy
 from steam_nb_api.ledet.ParameterSweep import *
 from nptdms import TdmsFile
 import nptdms
 from scipy import signal
 import csv
 from scipy.integrate import simps
+import time
+
 
 @dataclass
 class TDMSdata:
@@ -24,6 +27,7 @@ class TDMSdata:
     trigger_shoot: int = 0
     I_DCCT_MF: np.ndarray = np.array([])
     I_DCCT_HF: np.ndarray = np.array([])
+    U_PC: np.ndarray = np.array([])
     I_CLIQ: np.ndarray = np.array([])
     U_CLIQ: np.ndarray = np.array([])
     C_CLIQ: float = 0.0
@@ -33,6 +37,9 @@ class TDMSdata:
     QL1: float = 0.0
     QL2: float = 0.0
     Quenches: np.ndarray = np.array([])
+    R_crow: float = 0.0
+    R_circuit: float = 0.0
+    U_TH: float = 0.0
 
 @dataclass
 class SIMdata:
@@ -56,6 +63,11 @@ class SIMdata:
     T_adiabatic: np.ndarray = np.array([])
     el_connections: np.ndarray = np.array([])
 
+@dataclass
+class WrappedVoltages:
+    SIM: np.ndarray = np.array([])
+    TDMS: np.ndarray = np.array([])
+
 class QuenchPlanAnalysis():
     ## Domain: Helper and init function
     def __init__(self, FileNameTDMS, FileNameSIMData, beforePC = 0.4, verbose =  False):
@@ -65,7 +77,7 @@ class QuenchPlanAnalysis():
 
         self.MergedVoltages = {}
         self.MergedVoltagesPolarities = {}
-        self.CopyMVP = []
+        self.MergedVoltagesValues = {}
 
         self.SIMdata = SIMdata()
         self.SIMdata.FileName = FileNameSIMData
@@ -137,7 +149,9 @@ class QuenchPlanAnalysis():
         return idxQH.astype(int)
 
     def PrepareSimulations(self, PlanFile, RefFile, OutputDirectory, MagnetName = '', QHpairs = [[]], nQHseries = 2,
-                           t_QH_Fire = 0.002, t_CLIQ_Fire = 0.0005):
+                           t_QH_Fire = 0.002, t_CLIQ_Fire = 0.0005, t_EE_Fire = 0.008):
+        ## TODO Detection and Validation times!
+
         # Find MagnetName in RefFile Name
         if not MagnetName:
             RFile = RefFile.replace('\\', '//')
@@ -148,17 +162,28 @@ class QuenchPlanAnalysis():
         if self.verbose: print("MagnetName = ",MagnetName)
 
         ##  Prepare simulations
-        df = pd.read_excel(PlanFile)
+        df = pd.read_excel(PlanFile, engine = 'openpyxl')
+        xx = np.where(~df.iloc[0].notnull().to_numpy())[0]
+        df = df.drop(columns=df.columns.values[xx])
+
         SimNumbers = []
         for index, row in df.iterrows():
                 if index==0: continue
+                if np.isnan(df.iloc[index]['Current']): continue
+
                 a = ParametersLEDET()
                 a.readLEDETExcel(RefFile)
 
+                # Set temperature and tPC if available
+                if ~np.isnan(df.iloc[index]['T']):
+                    a.setAttribute("Inputs", "T00", df.iloc[index]['T'])
+                if ~np.isnan(df.iloc[index]['t_PC']):
+                    a.setAttribute("Inputs", "t_PC", df.iloc[index]['t_PC'])
+
                 ## 1. Adjust current in the circuit
-                a.setAttribute("Inputs", "I00", df.iloc[index]['Current'])
+                a.setAttribute("Inputs", "I00", df.iloc[index]['Current']*1000)
                 if df.iloc[index]['Trigger Type'] == 'SPA':
-                    [LUT, TimeLUT] = self.__constructLUT_SPA(df.iloc[index]['Ramp Rate'], df.iloc[index]['Current'],
+                    [LUT, TimeLUT] = self.__constructLUT_SPA(df.iloc[index]['Ramp Rate'], df.iloc[index]['Current']*1000,
                                                              a.Inputs.t_PC_LUT[0], a.Inputs.t_PC_LUT[1])
                     a.setAttribute('Inputs', 't_PC_LUT', TimeLUT)
                     if a.Options.time_vector_params[-1]<50:
@@ -167,16 +192,16 @@ class QuenchPlanAnalysis():
                                        self.__ExtendTimeVector(a.Options.time_vector_params, 50))
                 else:
                     LUT = a.Inputs.I_PC_LUT
-                    LUT[LUT != 0] = df.iloc[index]['Current']
+                    LUT[LUT != 0] = df.iloc[index]['Current']*1000
 
-                    if df.iloc[index]['Current'] < (a.Options.Iref / 10):
+                    if df.iloc[index]['Current']*1000 < (a.Options.Iref / 10):
                         print('Small current detected. I will refine the time vector.')
                         a.setAttribute('Options', 'time_vector_params',
                                        self.__RefineTimeVector(a.Options.time_vector_params))
-                    elif df.iloc[index]['Current'] < (a.Options.Iref / 4) and a.Options.time_vector_params[-1]<10:
+                    elif df.iloc[index]['Current']*1000 < (a.Options.Iref / 4) and a.Options.time_vector_params[-1]<10:
                         print("I assume a slow discharge. Extend time vector.")
-                        a.setAttribute('Options', 'time_vector_params',
-                                       self.__ExtendTimeVector(a.Options.time_vector_params, 10))
+                        #a.setAttribute('Options', 'time_vector_params',
+                         #              self.__ExtendTimeVector(a.Options.time_vector_params, 10))
 
                 a.setAttribute('Inputs', 'I_PC_LUT', LUT)
 
@@ -196,8 +221,10 @@ class QuenchPlanAnalysis():
 
                 ### 3. Adjust Energy Extraction
                 if df.iloc[index]['EE'] == df.iloc[index]['EE']:
-                    a.setAttribute("Inputs", "tEE", df.iloc[index]['Delay_EE'])
-                    a.setAttribute("Inputs", "R_EE", df.iloc[index]['R_EE'])
+                    if df.iloc[index]['Delay_EE'] < 0:
+                        a.setAttribute("Inputs", "tEE", df.iloc[index]['Delay_EE']/1000-t_EE_Fire )
+                    else: a.setAttribute("Inputs", "tEE", df.iloc[index]['Delay_EE']/1000+t_EE_Fire )
+                    a.setAttribute("Inputs", "R_EE_triggered", df.iloc[index]['R_EE'])
                 else:
                     a.setAttribute("Inputs", "tEE", 9999)
 
@@ -281,12 +308,8 @@ class QuenchPlanAnalysis():
         if len(f_CLIQ)<2: F_CYCLE = 0
         return f_CLIQ[F_CYCLE]
 
-    def _CalculateCLIQCapacitance(self, TimeFrame, I_CLIQ, U_CLIQ, kernel_size, I_tol = 50):
-        if len(I_CLIQ.shape)<2: I_CLIQ = np.array([I_CLIQ]).transpose()
-        if len(U_CLIQ.shape) < 2: U_CLIQ = np.array([U_CLIQ])
-        if len(TimeFrame.shape) < 2: TimeFrame = np.array([TimeFrame]).transpose()
-
-        filtered = signal.medfilt(U_CLIQ[0], kernel_size=kernel_size)
+    def _CalculateCLIQCapacitance(self, TimeFrame, I_CLIQ, U_CLIQ, kernel_size, I_tol = 50, Plot = 0):
+        filtered = signal.medfilt(U_CLIQ[:, 0], kernel_size=kernel_size)
         I_filt = signal.medfilt(I_CLIQ[:, 0], kernel_size=5)
 
         I_filt[abs(I_filt)<I_tol] = 0
@@ -295,13 +318,20 @@ class QuenchPlanAnalysis():
         IntI = simps(I_CLIQ[:cutIdx,0], TimeFrame[:cutIdx,0])
         U0 = filtered[0]
 
-        if len(filtered)-2000 > cutIdx: offIdx = -2000
-        elif len(filtered)-1000 > cutIdx: offIdx = -1000
-        else: offIdx = cutIdx
-        Uoffset = sum(filtered[offIdx:])/len(filtered[offIdx:])
-        Uend = filtered[cutIdx]-Uoffset
+        # if len(filtered)-2000 > cutIdx: offIdx = -2000
+        # elif len(filtered)-1000 > cutIdx: offIdx = -1000
+        # else: offIdx = cutIdx
+        # Uoffset = sum(filtered[offIdx:])/len(filtered[offIdx:])
+        Uend = filtered[cutIdx]
+        U0 = U0
         C = abs(IntI) /abs((Uend-U0))
         if self.verbose: print('Measured CLIQ capacitance: ', C * 1000, ' mF')
+
+        if Plot:
+            fig = plt.figure()
+            ax = fig.add_subplot(111)
+            ax.plot(TimeFrame, filtered)
+            ax.grid(True)
         return C
 
     def __calculateQuenchIntegral(self, time, current):
@@ -311,37 +341,31 @@ class QuenchPlanAnalysis():
 
     def _FindFiring(self, I, TimeFrame, dI_tol = 10):
         t_Start = []
-        if len(TimeFrame.shape) < 2: TimeFrame = np.array([TimeFrame]).transpose()
-        if len(I.shape)<2:
-            I = np.array([I]).transpose()
-            zeros = self.__zero_runs(I, Itol = dI_tol)
-            if not len(zeros) > 0 or abs(sum(I) / len(I)) < 0.5 or zeros[0][1] == len(I):
+        for i in range(I.shape[1]):
+            I_c = np.array([I[:,i]]).transpose()
+            zeros = self.__zero_runs(I_c, Itol = dI_tol)
+            if not len(zeros) > 0 or abs(sum(I_c ) / len(I_c )) < 0.5 or zeros[0][1] == len(I):
+                t_Start.append(np.nan)
                 if self.verbose: print('No signal found.')
-                return np.nan
-            t_Start = TimeFrame[zeros[0][1]]
-        else:
-            for i in range(I.shape[1]):
-                I_c = np.array([I[:,i]]).transpose()
-                zeros = self.__zero_runs(I_c, Itol = dI_tol)
-                if not len(zeros) > 0 or abs(sum(I_c ) / len(I_c )) < 0.5 or zeros[0][1] == len(I):
-                    t_Start.append(np.nan)
-                    if self.verbose: print('No signal found.')
-                    continue
-                t_Start.append(TimeFrame[zeros[0][1]][0])
+                continue
+            t_Start.append(TimeFrame[zeros[0][1]][0])
         return np.round(t_Start,7)
 
     def AvailableVoltageTaps(self):
         print('*** Available Voltage-taps ***')
         print(list(self.MergedVoltages.keys()))
 
-    def AvailableCoils(self):
-        print('*** Available Coils ***')
+    def AvailableCoils(self, Return = 0):
+        if not Return: print('*** Available Coils ***')
         current_coil = 0
+        coils = []
         for i in range(len(self.TDMSdata.VoltageVertices)):
             c_coil = self.TDMSdata.VoltageVertices[i][1]
             if c_coil != current_coil:
-                print(c_coil)
+                if not Return: print(c_coil)
                 current_coil = c_coil
+                coils.append(c_coil)
+        if Return: return coils
 
     def GetCoilStructure(self, Print = 0, Plot = 1):
         if Print:
@@ -380,6 +404,46 @@ class QuenchPlanAnalysis():
                   " to ", self.TDMSdata.VoltageVertices[i][4]+self.TDMSdata.VoltageVertices[i][5])
         return
 
+    def __gatherVoltageTaps(self):
+        Coil = self.AvailableCoils(Return = 1)
+        if not Coil:
+            if self.verbose: print('No coils found. Either provide Coil-structure first or check TDMS data.')
+        for c in range(len(Coil)):
+            Tap_Dict = {}
+            taps = []
+            for l in range(len(self.TDMSdata.VoltageVertices)):
+                if int(Coil[c]) == int(self.TDMSdata.VoltageVertices[l][1]):
+                    taps.append(self.TDMSdata.VoltageVertices[l][6])
+            for k in range(len(taps)):
+                TapName = taps[k]
+                realTapName = ''
+                for i in range(len(self.TDMSdata.VoltageVertices)):
+                    if self.TDMSdata.VoltageVertices[i][6] ==  TapName or self.TDMSdata.VoltageVertices[i][7] == TapName:
+                        realTapName = self.TDMSdata.VoltageVertices[i][8]
+                if not realTapName:
+                    print('Voltage Tap not found.')
+                    return
+                try:
+                    MVidx = list(self.MergedVoltages.keys()).index(realTapName)
+                except:
+                    try:
+                        _ = self.TDMSdata.groupMF[realTapName]
+                        print('Skip Voltage-tap ', realTapName, ' as no turns are bound to it [E1: MagnetEnds]')
+                        continue
+                    except:
+                        print('Voltage tap ', realTapName, ' was not found in TDMS data. [E2: TDMS]')
+                        continue
+                newV = WrappedVoltages()
+                newV.SIM = np.sum(self.SIMdata.Uturn_half_turns[self.MergedVoltages[realTapName].astype(int)-1], axis=0)*self.MergedVoltagesPolarities[realTapName]
+                newV.TDMS = self.TDMSdata.groupMF[realTapName].data[self.TDMSdata.trigger_shoot:]
+                # Correct for different length in Utaps and other channels
+                if len(newV.TDMS)<len(self.TDMSdata.TimeFrame_MF):
+                    len_diff = len(self.TDMSdata.TimeFrame_MF)-len(newV.TDMS)
+                    newV.TDMS = np.append(newV.TDMS, np.array([0]*len_diff))
+                Tap_Dict[realTapName] = newV
+            self.MergedVoltagesValues[Coil[c]] = Tap_Dict
+        return
+
     #############################################################################################################
     ## Domain: Handle SIM data Object
     def __CreateSIMDataObject(self):
@@ -401,7 +465,7 @@ class QuenchPlanAnalysis():
             self.SIMdata.I_CLIQ = self.SIMdata.I_CLIQ[trigger_shoot:]
             self.SIMdata.f_CLIQ = self.__FrequencyCLIQ(self.SIMdata.I_CLIQ, SIM=1)
             self.SIMdata.directionsCLIQ = np.array(f.get("directionCurrentCLIQ"))
-            U_CLIQ = np.array(f.get("Uc"))
+            U_CLIQ = np.array(f.get("Uc")).transpose()
             self.SIMdata.U_CLIQ = U_CLIQ[:][trigger_shoot:]
             if self.SIMdata.f_CLIQ > 0:
                 self.SIMdata.C_CLIQ = self._CalculateCLIQCapacitance(self.SIMdata.TimeFrame, self.SIMdata.I_CLIQ,
@@ -454,29 +518,35 @@ class QuenchPlanAnalysis():
         U_Heaters = []
         for channel in self.TDMSdata.groupHF.channels():
             if channel.name == 'IDCCT_HF': self.TDMSdata.I_DCCT_HF = channel.data
+            if len(self.TDMSdata.I_DCCT_HF.shape) < 2: self.TDMSdata.I_DCCT_HF = np.array([self.TDMSdata.I_DCCT_HF]).transpose()
         for channel in self.TDMSdata.groupMF.channels():
             if channel.name == 'IDCCT_HF':
                 self.TDMSdata.I_DCCT_MF = channel.data*1000
+                if len(self.TDMSdata.I_DCCT_MF.shape) < 2: self.TDMSdata.I_DCCT_MF = np.array([self.TDMSdata.I_DCCT_MF]).transpose()
                 self.TDMSdata.t_steps_MF = channel.properties['wf_samples']
                 self.TDMSdata.t_d_MF = channel.properties['wf_increment']
-            if channel.name == 'Trigger_PC' or channel.name == 'Trigger_HF':
-                trigger_data = channel.data
+            if channel.name == 'Trigger_PC' or channel.name == 'Trigger_HF': trigger_data = channel.data
             if channel.name == 'I Cliq' or channel.name == 'I_CLIQ':self.TDMSdata.I_CLIQ = channel.data
+            if len(self.TDMSdata.I_CLIQ.shape) < 2: self.TDMSdata.I_CLIQ = np.array([self.TDMSdata.I_CLIQ]).transpose()
             if channel.name == 'V Cliq' or channel.name == 'U_CLIQ': self.TDMSdata.U_CLIQ = channel.data
+            if len(self.TDMSdata.U_CLIQ.shape) < 2: self.TDMSdata.U_CLIQ = np.array([self.TDMSdata.U_CLIQ]).transpose()
+            if channel.name == 'Vpc': self.TDMSdata.U_PC = channel.data
+            if len(self.TDMSdata.U_PC.shape) < 2: self.TDMSdata.U_PC = np.array([self.TDMSdata.U_PC]).transpose()
             if channel.name.startswith("I_Heater") or channel.name.startswith("I(Alim__QH_"): I_Heaters.append(channel.name)
             if channel.name.startswith("U_Heater") or channel.name.startswith("U(Alim_QH_"): U_Heaters.append(channel.name)
         ## Set up Timeframe
-        self.TDMSdata.TimeFrame_MF = np.linspace(0,self.TDMSdata.t_steps_MF*self.TDMSdata.t_d_MF, int(self.TDMSdata.t_steps_MF))
+        self.TDMSdata.TimeFrame_MF = np.array([np.linspace(0,self.TDMSdata.t_steps_MF*self.TDMSdata.t_d_MF, int(self.TDMSdata.t_steps_MF))]).transpose()
         #self.TDMSdata.goupHF['IDCCT_HF'].time_track()
         trigger_shoot = np.gradient(trigger_data, self.TDMSdata.t_d_MF)
         trigger_shoot = np.where(abs(trigger_shoot)>self.TDMSdata.Tolerance)[0][0]-int(self.beforePC/self.TDMSdata.t_d_MF)
         self.TDMSdata.trigger_PC = int(self.beforePC/self.TDMSdata.t_d_MF)
         self.TDMSdata.trigger_shoot = trigger_shoot
-        self.TDMSdata.TimeFrame_MF = self.TDMSdata.TimeFrame_MF[trigger_shoot:]-self.TDMSdata.TimeFrame_MF[trigger_shoot]\
-                                     -self.TDMSdata.TimeFrame_MF[self.TDMSdata.trigger_PC+1]
-        self.TDMSdata.I_DCCT_MF = self.TDMSdata.I_DCCT_MF[trigger_shoot:]
-        self.TDMSdata.I_CLIQ = self.TDMSdata.I_CLIQ[trigger_shoot:]
-        self.TDMSdata.U_CLIQ = self.TDMSdata.U_CLIQ[trigger_shoot:]
+        self.TDMSdata.TimeFrame_MF = self.TDMSdata.TimeFrame_MF[trigger_shoot:, :]-self.TDMSdata.TimeFrame_MF[trigger_shoot, :]\
+                                     -self.TDMSdata.TimeFrame_MF[self.TDMSdata.trigger_PC+1, :]
+        self.TDMSdata.I_DCCT_MF = self.TDMSdata.I_DCCT_MF[trigger_shoot:, :]
+        self.TDMSdata.U_PC = self.TDMSdata.U_PC[trigger_shoot:, :]
+        self.TDMSdata.I_CLIQ = self.TDMSdata.I_CLIQ[trigger_shoot:, :]
+        self.TDMSdata.U_CLIQ = self.TDMSdata.U_CLIQ[trigger_shoot:, :]
         self.TDMSdata.f_CLIQ = self.__FrequencyCLIQ(self.TDMSdata.I_CLIQ, TDMS= 1)
         if self.TDMSdata.f_CLIQ > 0:
             self.TDMSdata.C_CLIQ = self._CalculateCLIQCapacitance(self.TDMSdata.TimeFrame_MF, self.TDMSdata.I_CLIQ,
@@ -549,7 +619,7 @@ class QuenchPlanAnalysis():
 
         return
 
-    def ProvideTurnsToCoilStructure(self, Coil = [], CoilCsv = ''):
+    def ProvideTurnsToCoilStructure(self, Coil = [], CoilCsv = '', Polarities = []):
         if Coil:
             if not self.MergedVoltages:
                 for i in range(len(Coil)):
@@ -565,13 +635,7 @@ class QuenchPlanAnalysis():
                             if old_key == self.TDMSdata.VoltageVertices[i][6] or  old_key == self.TDMSdata.VoltageVertices[i][7]:
                                 new_key = self.TDMSdata.VoltageVertices[i][8]
                                 self.MergedVoltages[new_key] = self.MergedVoltages.pop(old_key)
-                                # if int(self.TDMSdata.VoltageVertices[i][3]) > int(self.TDMSdata.VoltageVertices[i][5]):
-                                #     if self.TDMSdata.VoltageVertices[i][2] == 'I' and self.TDMSdata.VoltageVertices[i][4] == 'O':
-                                #         self.MergedVoltagesPolarities[new_key] = 1
-                                #     else: self.MergedVoltagesPolarities[new_key] = -1
-                                # else:
-                                #     self.MergedVoltagesPolarities[new_key] = 1
-                                self.MergedVoltagesPolarities[new_key] = self.CopyMVP[counter]
+                                self.MergedVoltagesPolarities[new_key] = Polarities[counter]
                                 counter = counter + 1
                                 found = 1
                                 break
@@ -579,6 +643,7 @@ class QuenchPlanAnalysis():
                             if self.verbose: print("Couldn't find voltage-tap: {} in TDMS.".format(old_key))
                             self.MergedVoltages.pop(old_key)
                             counter = counter + 1
+                self.__gatherVoltageTaps()
                 return
             else:
                 self.__CreateTDMSDataObject()
@@ -592,6 +657,7 @@ class QuenchPlanAnalysis():
                 turn_count = 0
                 current_dict = {}
                 Coil = []
+                Polarities = []
                 for row in csv_reader:
                     if not any(row): continue
                     if not row[0]: continue
@@ -613,9 +679,9 @@ class QuenchPlanAnalysis():
                     turns = self.SIMdata.el_connections[turn_count:turn_count+int(nTurns*2)]
                     current_dict[key_name] = turns[:,0]
                     turn_count = turn_count + int(nTurns*2)
-                    self.CopyMVP.append(float(row[4]))
+                    Polarities.append(float(row[4]))
                 Coil.append(current_dict)
-            self.ProvideTurnsToCoilStructure(Coil)
+            self.ProvideTurnsToCoilStructure(Coil, Polarities = Polarities)
             return
         else:
             print('Please provide either Coil-Dictionary or CSV-File containing the voltage-tap to turn binding')
@@ -644,7 +710,7 @@ class QuenchPlanAnalysis():
                     tSteps = int(ValidationTime / self.TDMSdata.t_d_MF)
                     QQidx = np.where(abs(Udata[Qidx[i]:Qidx[i] + tSteps]) > Vthr)[0]
                     if len(QQidx) >= (0.95*tSteps) and (Qidx[i]+tSteps)<(self.TDMSdata.trigger_shoot+self.TDMSdata.trigger_PC):
-                        tQuench = self.TDMSdata.TimeFrame_MF[Qidx[i]-self.TDMSdata.trigger_shoot]
+                        tQuench = self.TDMSdata.TimeFrame_MF[Qidx[i]-self.TDMSdata.trigger_shoot, :]
                         idx_tQuench = Qidx[i]-self.TDMSdata.trigger_shoot
                         found.append((key,  tQuench, idx_tQuench))
                         if Plot:
@@ -657,8 +723,8 @@ class QuenchPlanAnalysis():
                             ax.set_xlabel("Time, t [ms]")
                             ax.set_ylabel("Voltage [V]")
                             if idx_tQuench-tSteps<0: tSteps = idx_tQuench
-                            ax.plot(self.TDMSdata.TimeFrame_MF[idx_tQuench-tSteps:idx_tQuench+int(5*tSteps)]*1000, Udata[Qidx[i]-tSteps:Qidx[i]+int(5*tSteps)],'--')
-                            ax.axvline(self.TDMSdata.TimeFrame_MF[idx_tQuench]*1000,c='r')
+                            ax.plot(self.TDMSdata.TimeFrame_MF[idx_tQuench-tSteps:idx_tQuench+int(5*tSteps), :]*1000, Udata[Qidx[i]-tSteps:Qidx[i]+int(5*tSteps)],'--')
+                            ax.axvline(self.TDMSdata.TimeFrame_MF[idx_tQuench, :]*1000,c='r')
                             ax.axhline(Vthr, c='r', ls='--')
                             ax.legend(["$U_{Res,"+key+"}$","$t_{Quench}$","$U_{Threshold}$"])
                             # except:
@@ -693,7 +759,6 @@ class QuenchPlanAnalysis():
                     legend.append(tx)
                     cmap = cm.get_cmap('winter', len(found))
                     c = cmap(abs(found[j][1]*1000) / abs(found[0][1]*1000))
-                    c = np.array([c])
                     pos = ax.scatter(self.SIMdata.XY_mag_ave[0, idxTap.astype(int) - 1],
                                self.SIMdata.XY_mag_ave[1, idxTap.astype(int) - 1], s=10, c=c)
                     count = count + 1
@@ -705,6 +770,32 @@ class QuenchPlanAnalysis():
                 ax.legend(legend)
         self.TDMSdata.Quenches = np.array(found)
 
+    def __ExtractElectricalParameters(self):
+        I_th_tol = 5
+        # Extract correct, precise current level
+        I_init = sum(self.TDMSdata.I_DCCT_MF[:self.TDMSdata.trigger_PC, 0])/self.TDMSdata.trigger_PC
+        if self.verbose: print('Measured initial current: ', I_init)
+        # Extract CLIQ direction
+        # TODO
+        # Extract R_Circuit
+        U_init = sum(self.TDMSdata.U_PC[:self.TDMSdata.trigger_PC, 0])/self.TDMSdata.trigger_PC
+        self.TDMSdata.R_circuit = U_init/I_init
+        if self.verbose: print('Measured lead resistance: ',self.TDMSdata.R_circuit)
+
+        # Find Ud_crowbar
+        I_th = signal.medfilt(self.TDMSdata.I_DCCT_MF[:,0], kernel_size = 101)
+        I_th[abs(I_th) < I_th_tol] = 0
+        ranges = self.__zero_runs(I_th)[-1][0]
+
+        idx_add = int(0.05/self.TDMSdata.t_d_MF)
+        self.TDMSdata.Ud_crow = abs(np.sum(self.TDMSdata.U_PC[ranges:ranges+idx_add,0])/len(self.TDMSdata.U_PC[ranges:ranges+idx_add,0]))
+        if self.verbose: print('Measured Ud_crowbar: ', self.TDMSdata.Ud_crow)
+
+        # R_crowbar
+
+
+        # R_CLIQ/R_QH
+
     def ExtractFeatures(self):
         ## Extract Timing features:
         print('Timing')
@@ -713,7 +804,7 @@ class QuenchPlanAnalysis():
         for i in range((dIdt_Coil.shape[1])):
             dIdt_Coil[:,i] = np.gradient(self.SIMdata.I_CoilSections[:,i], self.SIMdata.TimeFrame[:,0])
         print('t_PC_Sim: ', self._FindFiring(dIdt_Coil, self.SIMdata.TimeFrame))
-        dIdt_Coil = np.gradient(signal.medfilt(self.TDMSdata.I_DCCT_MF, kernel_size = 17), self.TDMSdata.TimeFrame_MF)
+        dIdt_Coil = np.array([np.gradient(signal.medfilt(self.TDMSdata.I_DCCT_MF[:,0], kernel_size = 17), self.TDMSdata.TimeFrame_MF[:,0])]).transpose()
         print('t_PC_Meas: ', self._FindFiring(dIdt_Coil, self.TDMSdata.TimeFrame_MF, dI_tol = 10000))
         # # Extract t_QH
         dIdt_QH = np.zeros((self.SIMdata.I_QH.shape))
@@ -722,22 +813,21 @@ class QuenchPlanAnalysis():
         print('t_QH_Sim: ', self._FindFiring(dIdt_QH.transpose(), self.SIMdata.TimeFrame))
         dIdt_QH = np.zeros((self.TDMSdata.I_QH.shape))
         for i in range((dIdt_QH.shape[0])):
-            dIdt_QH[i,:] = np.gradient(signal.medfilt(self.TDMSdata.I_QH[i,:], kernel_size = 11), self.TDMSdata.TimeFrame_MF)
+            dIdt_QH[i,:] = np.gradient(signal.medfilt(self.TDMSdata.I_QH[i,:], kernel_size = 11), self.TDMSdata.TimeFrame_MF[:,0])
         print('t_QH_Meas: ', self._FindFiring(dIdt_QH.transpose(), self.TDMSdata.TimeFrame_MF, dI_tol=1000))
         # # # Extract t_CLIQ
-        dIdt_CLIQ = np.gradient(self.SIMdata.I_CLIQ[:,0], self.SIMdata.TimeFrame[:,0])
+        dIdt_CLIQ = np.array([np.gradient(self.SIMdata.I_CLIQ[:,0], self.SIMdata.TimeFrame[:,0])]).transpose()
         print('t_CLIQ_Sim: ', self._FindFiring(dIdt_CLIQ, self.SIMdata.TimeFrame))
-        dIdt_CLIQ = np.gradient(signal.medfilt(self.TDMSdata.I_CLIQ, kernel_size = 11), self.TDMSdata.TimeFrame_MF)
+        dIdt_CLIQ = np.array([np.gradient(signal.medfilt(self.TDMSdata.I_CLIQ[:,0], kernel_size = 11), self.TDMSdata.TimeFrame_MF[:,0])]).transpose()
         print('t_CLIQ_Meas: ', self._FindFiring(dIdt_CLIQ, self.TDMSdata.TimeFrame_MF, dI_tol = 5000))
         # # # Extract t_EE
-        # # Extract t_Quench [if applicable]
+        ## TODO
+
+        ## Extract t_Quench [if applicable]
+        # self.FindQuenchTDMS(0.1, 0.007)
 
         ## Extract electrical parameters
-            # Extract correct, precise current level
-            # Done # Extract f and C_CLIQ
-            # Extract CLIQ direction
-            # Extract R_Circuit
-            # What about R_Crowbar/Ud_Crowbar/R_CLIQ/R_QH
+        self.__ExtractElectricalParameters()
 
         ## Extract further properties
             #To be seen if in this function
@@ -763,8 +853,8 @@ class QuenchPlanAnalysis():
         if TDMS:
             if not self.TDMSdata.Quenches.size > 0: self.FindQuenchTDMS(0.1, 0.007, Plot =0)
             idxQuench_TDMS = int(self.TDMSdata.Quenches[0][2])
-            self.TDMSdata.QL1 = self.__calculateQuenchIntegral(self.TDMSdata.TimeFrame_MF[idxQuench_TDMS:],
-                                                               self.TDMSdata.I_DCCT_MF[idxQuench_TDMS:])
+            self.TDMSdata.QL1 = self.__calculateQuenchIntegral(self.TDMSdata.TimeFrame_MF[idxQuench_TDMS:,0],
+                                                               self.TDMSdata.I_DCCT_MF[idxQuench_TDMS:,0])
         ## QL Sim Data
         if SIM:
             idxQuench_SIM = self.__FindQuenchSIM(self.SIMdata.Quenches[1])
@@ -780,8 +870,8 @@ class QuenchPlanAnalysis():
         ## QL Integral 2  - from tPC
         ## QL Measured Data
         if TDMS:
-            self.TDMSdata.QL2 = self.__calculateQuenchIntegral(self.TDMSdata.TimeFrame_MF[self.TDMSdata.trigger_PC:],
-                                                               self.TDMSdata.I_DCCT_MF[self.TDMSdata.trigger_PC:])
+            self.TDMSdata.QL2 = self.__calculateQuenchIntegral(self.TDMSdata.TimeFrame_MF[self.TDMSdata.trigger_PC:,0],
+                                                               self.TDMSdata.I_DCCT_MF[self.TDMSdata.trigger_PC:,0])
         ## QL Sim Data
         if SIM:
             if abs(sum(self.SIMdata.I_CLIQ)) > 0:
@@ -820,46 +910,30 @@ class QuenchPlanAnalysis():
         return [self.TDMSdata, self.SIMdata]
 
     def PlotVoltageTap(self, Coil = '', TapName = ''):
+        if not self.MergedVoltagesValues: self.__gatherVoltageTaps()
         if not Coil:
             if TapName:
                 title = TapName
-                Coil = [TapName]
+                Coil_List = [TapName]
+                for i in list(self.MergedVoltagesValues.keys()):
+                    if TapName in list(self.MergedVoltagesValues[i]): Coil = i
             else:
                 print('Please provide tap-name or coil to plot.')
                 return
         elif Coil:
             title = 'Coil '+str(Coil)
-            newC = []
-            for l in range(len(self.TDMSdata.VoltageVertices)):
-                if Coil == int(self.TDMSdata.VoltageVertices[l][1]):
-                    newC.append(self.TDMSdata.VoltageVertices[l][6])
-            Coil = newC
+            if not isinstance(Coil, str): Coil = str(Coil)
+            Coil_List = list(self.MergedVoltagesValues[Coil].keys())
+
         fig = plt.figure(figsize=(12,12))
         ax = fig.add_subplot(111)
         ax.grid(True)
         leg = []
-        for c in range(len(Coil)):
-            TapName = Coil[c]
-            realTapName = ''
-            for i in range(len(self.TDMSdata.VoltageVertices)):
-                if self.TDMSdata.VoltageVertices[i][6] ==  TapName or self.TDMSdata.VoltageVertices[i][7] == TapName:
-                    realTapName = self.TDMSdata.VoltageVertices[i][8]
-            if not realTapName:
-                print('Voltage Tap not found.')
-                return
-            try:
-                MVidx = list(self.MergedVoltages.keys()).index(realTapName)
-            except:
-                try:
-                    _ = self.TDMSdata.groupMF[realTapName]
-                    print('Skip Voltage-tap ', realTapName, ' as no turns are bound to it [E1: MagnetEnds]')
-                    continue
-                except:
-                    print('Voltage tap ', realTapName, ' was not found in TDMS data.')
-                    continue
-            LEDET_U = np.sum(self.SIMdata.Uturn_half_turns[self.MergedVoltages[realTapName].astype(int)-1], axis=0)*self.MergedVoltagesPolarities[realTapName]
-            TDMS_U = self.TDMSdata.groupMF[realTapName].data[self.TDMSdata.trigger_shoot:]
-            print_rTN = realTapName.replace('_', '-')
+        for c in range(len(Coil_List)):
+            TapName = Coil_List[c]
+            LEDET_U = self.MergedVoltagesValues[Coil][TapName].SIM
+            TDMS_U = self.MergedVoltagesValues[Coil][TapName].TDMS
+            print_rTN = TapName.replace('_', '-')
             ax.plot(self.TDMSdata.TimeFrame_MF, TDMS_U, '-', color= self.Colors[c])
             leg.append('$U_{Meas,' + print_rTN + '}$')
             ax.plot(self.SIMdata.TimeFrame, LEDET_U, '--', color= self.Colors[c])
@@ -871,6 +945,53 @@ class QuenchPlanAnalysis():
         ax.set_xlabel("Time [s]", fontsize=20)
         return
 
-    def SaveToCSV(self):
-        # TODO
+    def SaveToCSV(self, Coil = '', TapName = ''):
+        if not self.MergedVoltagesValues: self.__gatherVoltageTaps()
+        if not Coil:
+            if TapName:
+                Coil_List = [TapName]
+                for i in list(self.MergedVoltagesValues.keys()):
+                    if TapName in list(self.MergedVoltagesValues[i]): Coil = i
+            else:
+                print('Please provide tap-name or coil to plot.')
+                return
+        elif Coil:
+            title = 'Coil ' + str(Coil)
+            if not isinstance(Coil, str): Coil = str(Coil)
+            Coil_List = list(self.MergedVoltagesValues[Coil].keys())
+
+        Voltages_Sim = []
+        Voltages_Tdms  = []
+        Tap_Names = []
+        Unit_Names = []
+        for c in range(len(Coil_List)):
+            TapName = Coil_List[c]
+            LEDET_U = self.MergedVoltagesValues[Coil][TapName].SIM
+            TDMS_U = self.MergedVoltagesValues[Coil][TapName].TDMS
+            if c==0:
+                Voltages_Sim = LEDET_U
+                Voltages_Tdms = TDMS_U
+                csv_length = max(Voltages_Tdms.shape[0], Voltages_Sim.shape[0])
+            else:
+                Voltages_Sim = np.vstack((Voltages_Sim, LEDET_U))
+                Voltages_Tdms = np.vstack((Voltages_Tdms, TDMS_U))
+            print_rTN = TapName.replace('_', '-')
+            Tap_Names = Tap_Names + [print_rTN, '','','','']
+            Unit_Names = Unit_Names +['t_MEAS','U_'+print_rTN, 't_SIM', 'U_SIM', '']
+
+        with open('Voltages_'+Coil+'.csv', mode='w', newline='') as file:
+            writer = csv.writer(file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
+            writer.writerow(Tap_Names)
+            writer.writerow(Unit_Names)
+            for i in range(csv_length):
+                row = []
+                if i>Voltages_Sim.shape[1]-1: i1 = Voltages_Sim.shape[1]-1
+                else: i1 = i
+                if i>Voltages_Tdms.shape[1]-1: i2 = Voltages_Tdms.shape[1]-1
+                else: i2 = i
+
+                for k in range(Voltages_Sim.shape[0]):
+                    row = row + [self.SIMdata.TimeFrame[i1][0],  Voltages_Sim[k, i1], self.TDMSdata.TimeFrame_MF[i2],
+                                 Voltages_Tdms[k, i2], '']
+                writer.writerow(row)
         return
