@@ -9,7 +9,7 @@ from steam_nb_api.ledet.QuenchAnalysis import QuenchPlanAnalysis
 from steam_nb_api.ledet.AutomaticSweep import AutomaticSweep
 import time
 
-from steam_nb_api.utils.MP3_simulations import *
+from steam_nb_api.utils.STEAMLib_simulations import *
 
 testfile1 = "C:\cernbox\SWAN_projects\steam-notebooks\steam-ledet-input\MBRC\MBRC_0.xlsx"
 testfile2 = "C:\cernbox\SWAN_projects\steam-notebooks\steam-ledet-input\MBRB\MBRB_0.xlsx"
