@@ -189,7 +189,8 @@ class ParametersCOSIM:
     def makeGenericIOPortFiles(self, CoilSections, ModelFolder, CoSimFolder: str, PSpiceExecutable: str, LEDETExecutable: str, verbose: bool = False,
                                t_0 = [0, 2e-5], t_end = [2e-5, 0.5], t_step_max = [[1e-5, 1e-5], [1e-5, 1e-5]],
                                relTolerance = [1e-4, None], absTolerance = [1, None], executionOrder = [1, 2],
-                               executeCleanRun = [True, True], N_LEDET = 1, QuenchMagnet = 0, SimulationNumber = [0], DistinctMagnets = 1, PSPICEinitialConditions = []):
+                               executeCleanRun = [True, True], N_LEDET = 1, QuenchMagnet = 0, SimulationNumber = [0], DistinctMagnets = 1, PSPICEinitialConditions = [],
+                               convergenceElement = 'x_mag_1'):
 
         if ModelFolder.endswith('//'): ModelFolder.replace('//','\\')
         elif not ModelFolder.endswith('\\'): ModelFolder = ModelFolder + '\\'
@@ -213,9 +214,9 @@ class ParametersCOSIM:
 
         # Variables used for convergence, for now its the first magnet current and the LEDET Voltages
         if N_LEDET == 1:
-            convergenceVariables = ["I(x_mag_1." + Components[0] + "_1)", "U_inductive_dynamic_" + Components[1]+"_1"]
+            convergenceVariables = ["I(" + convergenceElement + "." + Components[0] + "_1)", "U_inductive_dynamic_" + Components[1]+"_1"]
         else:
-            convergenceVariables = ["I(x_mag_1." + Components[0] + "_1)", "U_inductive_dynamic_" + Components[1]+"_1"]
+            convergenceVariables = ["I(" + convergenceElement + "." + Components[0] + "_1)", "U_inductive_dynamic_" + Components[1]+"_1"]
             for i in range(2,N_LEDET+1):
                 convergenceVariables.append("U_inductive_dynamic_" + Components[1]+"_1")
 
